// Copyright (C) 2017 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#include "qt3dquickscene2d_logging_p.h"

QT_BEGIN_NAMESPACE

namespace Qt3DRender {
namespace Quick {

Q_LOGGING_CATEGORY(Scene2D, "Qt3D.Scene2D", QtWarningMsg)

} // Quick
} // Qt3DRender

QT_END_NAMESPACE
