// Copyright (C) 2022 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#include <Qt3DAnimation/qanimationcallback.h>

QT_BEGIN_NAMESPACE

namespace Qt3DAnimation {

QAnimationCallback::~QAnimationCallback()
    = default;

} // namespace Qt3DAnimation

QT_END_NAMESPACE
