// Copyright (C) 2017 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#ifndef QT3D_QUICKWINDOWLOGGING_P_H
#define QT3D_QUICKWINDOWLOGGING_P_H

//
//  W A R N I N G
//  -------------
//
// This file is not part of the Qt API.  It exists purely as an
// implementation detail.  This header file may change from version to
// version without notice, or even be removed.
//
// We mean it.
//

#include <QLoggingCategory>
#include <private/qglobal_p.h>

QT_BEGIN_NAMESPACE

namespace Qt3DExtras {

Q_DECLARE_LOGGING_CATEGORY(QuickWindow)

} // Qt3DRender

QT_END_NAMESPACE

#endif // QT3D_QUICKWINDOWLOGGING_P_H
