// Copyright (C) 2016 Klaralvdalens Datakonsult AB (KDAB).
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#include "abstractactioninput_p.h"

QT_BEGIN_NAMESPACE

namespace Qt3DInput {

namespace Input {

AbstractActionInput::AbstractActionInput()
    : BackendNode()
{
}

} // namespace Input

} // namespace Qt3DInput

QT_END_NAMESPACE

