// Copyright (C) 2015 Klaralvdalens Datakonsult AB (KDAB).
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#include "boundingsphere_p.h"

QT_BEGIN_NAMESPACE

namespace Qt3DRender {
namespace RayCasting {

BoundingSphere::BoundingSphere()
{
}

BoundingSphere::~BoundingSphere()
{
}

} // namespace RayCasting
} // namespace Qt3DRender

QT_END_NAMESPACE
