// Copyright (C) 2017 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#include "qt3dquickwindowlogging_p.h"

QT_BEGIN_NAMESPACE

namespace Qt3DExtras {

Q_LOGGING_CATEGORY(QuickWindow, "Qt3D.QuickWindow", QtWarningMsg)

} // Qt3DExtras

QT_END_NAMESPACE
